/*
 * names.c: This here is used to maintain a list of all the people currently
 * on your channel.  Seems to work 
 *
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */


#include "irc.h"

#include "ircaux.h"
#include "names.h"
#include "flood.h"
#include "window.h"
#include "screen.h"
#include "server.h"
#include "lastlog.h"
#include "list.h"
#include "output.h"
#include "userlist.h"
#include "timer.h"
#include "hook.h"
#include "pluto.h"
#include "parse.h"
#include "whowas.h"
#include "misc.h"

extern int in_on_who;
extern ChannelList default_statchan;
extern AJoinList *ajoin_list;

static	char	mode_str[] = "iklmnpst";

static	void	add_to_mode_list _((char *, int, char *));
static	void	check_mode_list_join _((char *, int));
#if 0
static	void	free_channel _((ChannelList **));
#endif
static	void	show_channel _((ChannelList *));
static	void	clear_channel _((ChannelList *));
static	char	*recreate_mode _((ChannelList *));
static	int	decifer_mode _((char *, char *, ChannelList **, u_long *, char *, NickList **, char **));
	char	*channel_key _((char *));

extern  char * function_getkey _((unsigned char *));
static
struct modelist
{
	char	*chan;
	int	server;
	char	*mode;
	struct modelist *next;
}	*mode_list = NULL;


/* clear_channel: erases all entries in a nick list for the given channel */
static	void
clear_channel(chan)
	ChannelList *chan;
{
	NickList *tmp,
		*next;

	for (tmp = chan->nicks; tmp; tmp = next)
	{
		next = tmp->next;
		new_free(&(tmp->host));
		new_free(&(tmp->nick));
		new_free(&(tmp->server));
		new_free((char **)&tmp);
	}
	chan->nicks = NULL;
}

extern	ChannelList *
lookup_channel(channel, server, unlink)
	char	*channel;
	int	server;
	int	unlink;
{
	ChannelList	*chan = server_list[server].chan_list,
			*last = NULL;

	if (server == -1)
		server = primary_server;

	while (chan)
	{
		if (chan->server == server && !my_stricmp(chan->channel, channel))
		{
			if (unlink == CHAN_UNLINK)
			{
				if (last)
					last->next = chan->next;
				else
					server_list[server].chan_list = chan->next;
			}
			break;
		}
		last = chan;
		chan = chan->next;
	}
	return chan;
}

#ifdef __STDC__
extern void set_waiting_channel (int i)
#else
extern void set_waiting_channel(i)
	int	i;
#endif
{
	Window	*tmp;
	int	flag = 1;

	while (NULL != (tmp = traverse_all_windows(&flag)))
		if (tmp->server == i && tmp->current_channel)
		{
			if (tmp->bind_channel)
				tmp->waiting_channel = tmp->current_channel;
			tmp->current_channel = NULL;
		}
}

/* if the user is on the given channel, it returns 1. */
#ifdef __STDC__
extern int im_on_channel (char *channel)
#else
extern int im_on_channel (channel)
char *channel;
#endif
{
	return (lookup_channel(channel, from_server, 0)) ? 1 : 0;
}

/*
 * add_channel: adds the named channel to the channel list.  If the channel
 * is already in the list, then the channel gets cleaned, and ready for use
 * again.   The added channel becomes the current channel as well.
 */
void
add_channel(channel, server)
	char	*channel;
	int	server;
{
	ChannelList *new;
	WhowasChanList *whowaschan;
	
	if ((new = lookup_channel(channel, server, CHAN_NOUNLINK)) != NULL)
	{
		new->connected = 1;
		new->server = server;
		new->window = curr_scr_win;
		malloc_strcpy(&(new->channel), channel);
		clear_channel(new);
	}
	else
	{
		if (!(whowaschan = check_whowas_chan_buffer(channel, 1)))
		{
			new = (ChannelList *) new_malloc(sizeof(ChannelList));
			memcpy(new, &default_statchan, sizeof(ChannelList));
			new->connected = 1;
			get_time(&new->channel_create);
			malloc_strcpy(&(new->channel), channel);
		}
		else
		{
			new = whowaschan->channellist;
			new_free(&whowaschan->channel);
			new_free((char **)&whowaschan);
			new_free(&(new->key));
			new->mode = 0;
			clear_channel(new);
		}
		new->server = server;
		new->window = curr_scr_win;
		new->flags.got_modes = new->flags.got_who = new->flags.got_bans = 1;
		get_time(&new->join_time);
		add_to_list((List **)&(server_list[server].chan_list), (List *)new);
	}

	if (!is_current_channel(channel, server, 0))
	{
		int	flag = 1;
		Window	*tmp;

		while ((tmp = traverse_all_windows(&flag)) != NULL)
		{
/*
                        if (!tmp->waiting_channel && !tmp->bind_channel)
                                 continue;
*/
                        if (tmp->server != from_server)
                                continue;
                        if (tmp->bind_channel && !my_stricmp(tmp->bind_channel, channel))
                        {
                                set_channel_by_refnum(tmp->refnum, channel);
                                new->window = tmp;
                                update_all_windows();
                                return;
                        }
                        if (tmp->waiting_channel && !my_stricmp(tmp->waiting_channel, channel))
                        {
				set_channel_by_refnum(tmp->refnum, channel);
				new->window = tmp;
				new_free(&tmp->waiting_channel);
				update_all_windows();
				return;
			}
                        set_channel_by_refnum(tmp->refnum, channel);
                        new->window = tmp;
                        update_all_windows();
                        return;
		}
		set_channel_by_refnum(0, channel);
		new->window = curr_scr_win;
	}
	update_all_windows();
}

/*
 * add_to_channel: adds the given nickname to the given channel.  If the
 * nickname is already on the channel, nothing happens.  If the channel is
 * not on the channel list, nothing happens (although perhaps the channel
 * should be addded to the list?  but this should never happen) 
 */
void
add_to_channel(char *channel, char *nick, int server, int oper, int voice, char *userhost, char *server1, char *away)
{
	NickList *new = NULL;
	ChannelList *chan = NULL;
	WhowasList *whowas;

	int	ischop = oper;
	
	if ((chan = lookup_channel(channel, server, CHAN_NOUNLINK)))
	{
		if (*nick == '+')
			nick++;
		if (*nick == '@')
		{
			nick++;
			if (my_stricmp(nick, get_server_nickname(server)) == 0)
			{
				if (!list_lookup((List **) chan, nick, !USE_WILDCARDS, !REMOVE_FROM_LIST))
					check_mode_list_join(channel, server);
				chan->chop = 1;
			}
			ischop = 1;
		}
		if (!(new = (NickList *) find_in_list((List **)&(chan->nicks), nick, 0)))
		{
			if (!(whowas = check_whowas_buffer(nick, userhost, channel, 1)))
			{
				new = (NickList *) new_malloc(sizeof(NickList));
				new->idle_time = new->kicktime = new->doptime = new->nicktime = new->floodtime = time(NULL);
				new->nickcount = new->dopcount = new->kickcount = new->floodcount = 0;
				malloc_strcpy(&(new->nick), nick);
				new->userlist = nickinuser(nick, userhost);
				new->shitlist = nickinshit(nick, userhost);
				new->botlist = nickinbot(nick, userhost);		
			} 
			else
			{
				new = whowas->nicklist;
#if 0				
				if (server2 && whowas->server2)
				{
					if ((my_stricmp(server2, whowas->server2)))
					{
						yell("DEBUG *** servers don't match !updating");
						malloc_strcpy(&new->server, server2);
					}
					else
						malloc_strcpy(&new->server, whowas->server2);
				}
#endif
				new_free(&whowas->channel);
				new_free(&whowas->server1);
				new_free(&whowas->server2);
				new_free((char **)&whowas);
				malloc_strcpy(&(new->nick), nick);
			}
			add_to_list((List **) &(chan->nicks), (List *) new);
			update_stats(JOINLIST, chan->channel, nick, chan, 0);
		}
		new->chanop = ischop;
		new->voice = voice;
		if (away)
		{
			new->away = *away;
			new->ircop = *(away+1) == '*';
		}
		if (server1)
			malloc_strcpy(&new->server, server1);
		if (userhost)
			malloc_strcpy(&new->host, userhost);
	}
}


/*
 * recreate_mode: converts the bitmap representation of a channels mode into
 * a string 
 *
 * This malloces what it returns, but nothing that calls this function
 * is expecting to have to free anything.  Therefore, this function
 * should not malloc what it returns.  (hop)
 *
 * but this leads to horrible race conditions, so we add a bit to
 * the channel structure, and cache the string value of mode, and
 * the u_long value of the cached string, so that each channel only
 * has one copy of the string.  -mrg, june '94.
 */
static	char	*
recreate_mode(chan)
	ChannelList *chan;
{
	int	mode_pos = 0,
		mode;
	static	char	*s;
	char	buffer[BIG_BUFFER_SIZE + 1];
	/* first check if cached string value is ok */

	if (chan->mode == chan->i_mode)
		return (chan->s_mode);

	chan->i_mode = chan->mode;
	buffer[0] = '\0';
	s = buffer;
	mode = chan->mode;
	while (mode)
	{
		if (mode % 2)
			*s++ = mode_str[mode_pos];
		mode /= 2;
		mode_pos++;
	}
	if (chan->key && *chan->key)
	{
		*s++ = ' ';
		strcpy(s, chan->key);
		s += strlen(chan->key);
	}
	if (chan->limit)
		sprintf(s, " %d", chan->limit);
	else
		*s = '\0';

	malloc_strcpy(&chan->s_mode, buffer);
	return chan->s_mode;
}

/*
 * decifer_mode: This will figure out the mode string as returned by mode
 * commands and convert that mode string into a one byte bit map of modes 
 */
static	int
decifer_mode(from, mode_string, channel, mode, chop, nicks, key)
	char	*from;
	char	*mode_string;
	ChannelList **channel;
	u_long	*mode;
	char	*chop;
	NickList **nicks;
	char	**key;
{

	char	*limit = 0;
	char	*person;
	int	add = 0;
	int	limit_set = 0;
	int	limit_reset = 0;
	int	splitter = 0;
	char	*rest,
		*the_key;
	
	NickList *ThisNick = NULL;
	BanList *new;
	unsigned char	value = 0;
	
	if (!(mode_string = next_arg(mode_string, &rest)))
		return -1;
	splitter = match("*.*.*", from);
	for (; *mode_string; mode_string++)
	{
		switch (*mode_string)
		{
		case '+':
			add = 1;
			value = 0;
			break;
		case '-':
			add = 0;
			value = 0;
			break;
		case 'p':
			value = MODE_PRIVATE;
			break;
		case 'l':
			value = MODE_LIMIT;
			if (add)
			{
				limit_set = 1;
				if (!(limit = next_arg(rest, &rest)))
					limit = empty_string;
				else if (0 == strncmp(limit, zero, 1))
					limit_reset = 1, limit_set = 0, add = 0;
			}
			else
				limit_reset = 1;
			break;
		case 't':
			value = MODE_TOPIC;
			break;
		case 'i':
			value = MODE_INVITE;
			break;
		case 'n':
			value = MODE_MSGS;
			break;
		case 's':
			value = MODE_SECRET;
			break;
		case 'm':
			value = MODE_MODERATED;
			break;
		case 'o':
			if ((person = next_arg(rest, &rest)) && 
			    !my_stricmp(person, get_server_nickname(from_server)))
				*chop = add;
			ThisNick = (NickList *) list_lookup((List **)nicks, person,
                        	!USE_WILDCARDS, !REMOVE_FROM_LIST);
			if (ThisNick)
				ThisNick->chanop=add;
			update_stats(add ? MODEOPLIST: MODEDEOPLIST, (*channel)->channel, from, *channel, splitter);
			if (add == 0 && my_stricmp(from, get_server_nickname(from_server)))
				check_prot(from, person, *channel);
			if (add && my_stricmp(from, get_server_nickname(from_server)))
				check_hack(person, *channel, ThisNick, from);
			break;
			
		case 'k':
			value = MODE_KEY;
			the_key = next_arg(rest, &rest);
			if (add)
				malloc_strcpy(key, the_key);
			else
				new_free(key);
			break;	
		case 'v':
			if ((person = next_arg(rest, &rest)))
			{
				ThisNick = (NickList *) list_lookup((List **)nicks, person,
        	                	!USE_WILDCARDS, !REMOVE_FROM_LIST);
				if (ThisNick)
					ThisNick->voice=add;
			}
			break;
		case 'b':
			person = next_arg(rest, &rest);
			if (!person || !*person)
				break;
			update_stats(add?MODEBANLIST:MODEUNBANLIST, (*channel)->channel, from, *channel, splitter);
			if (add)
			{
				if (!(new = (BanList *)find_in_list((List **)&(*channel)->bans, person, 0)))
				{
					new = (BanList *) new_malloc(sizeof(BanList));
					malloc_strcpy(&new->ban, person);
					add_to_list((List **)&(*channel)->bans, (List *)new);
				} 
				new->sent_unban = 0;
				malloc_strcpy(&new->setby, from?from:get_server_name(from_server));
				new->time = time(NULL);
				check_prot(from, person, *channel);
			} 
			else
			{
				if ((new = (BanList *)find_in_list((List **)&(*channel)->bans, person, 0)))
				{
					new->sent_unban = 1;
					new->sent_unban_time = time(NULL);
				}
			}
/*			check_shit(person);*/
			break;
		}
		if (add)
			*mode |= value;
		else
			*mode &= ~value;
	}
	if (limit_set)
		return (atoi(limit));
	else if (limit_reset)
		return(0);
	else
		return(-1);
}

/*
 * get_channel_mode: returns the current mode string for the given channel
 */
char	*
get_channel_mode(channel, server)
	char	*channel;
	int	server;
{
	ChannelList *tmp;

	if ((tmp = lookup_channel(channel, server, CHAN_NOUNLINK)))
		return recreate_mode(tmp);
	return empty_string;
}

/*
 * update_channel_mode: This will modify the mode for the given channel
 * according the the new mode given.  
 */
void
update_channel_mode(from, channel, server, mode)
	char	*from;
	char	*channel;
	int	server;
	char	*mode;
{
	ChannelList *tmp;
	int	limit;

	if ((tmp = lookup_channel(channel, server, CHAN_NOUNLINK)) &&
			(limit = decifer_mode(from, mode, &(tmp), &(tmp->mode), &(tmp->chop), &(tmp->nicks), &(tmp->key))) != -1)
		tmp->limit = limit;
}

/*
 * is_channel_mode: returns the logical AND of the given mode with the
 * channels mode.  Useful for testing a channels mode 
 */
int
is_channel_mode(channel, mode, server_index)
	char	*channel;
	int	mode;
	int	server_index;
{
	ChannelList *tmp;

	if ((tmp = lookup_channel(channel, server_index, CHAN_NOUNLINK)))
		return (tmp->mode & mode);
	return 0;
}

#if 0
static	void
free_channel(channel)
	ChannelList **channel;
{
	clear_channel(*channel);
	new_free(&(*channel)->channel);
	new_free(&(*channel)->key);
	new_free((char **)&(*channel));
}
#endif

static	void
clear_bans(channel)
	ChannelList *channel;
{
	BanList *bans, 
		*next;
	if (!channel || !channel->bans)
		return;
	for (bans = channel->bans; bans; bans = next)
	{
		next = bans->next;
		new_free(&bans->setby);
		new_free(&bans->ban);
		new_free((char **)&bans);
	}
	channel->bans = NULL;
}

/*
 * remove_channel: removes the named channel from the
 * server_list[server].chan_list.  If the channel is not on the
 * server_list[server].chan_list, nothing happens.  If the channel was
 * the current channel, this will select the top of the
 * server_list[server].chan_list to be the current_channel, or 0 if the
 * list is empty. 
 */
void
remove_channel(channel, server)
	char	*channel;
	int	server;
{
	ChannelList *tmp;
	NickList *Nick;
	if (channel)
	{
		if ((tmp = lookup_channel(channel, server, CHAN_UNLINK)))
		{
			clear_bans(tmp);
			for (Nick = tmp->nicks; Nick; Nick = Nick->next)
				add_to_whowas_buffer(Nick, tmp->channel, Nick->server, get_server_name(server));
			tmp->nicks = NULL;
			add_to_whowas_chan_buffer(tmp);
		}

		if (is_current_channel(channel, server, -1))
			switch_channels(0,NULL);
		
	/* 1 would reference a window, which we don't want -Soleil */
	}
	else
	{
		ChannelList *next;

		for (tmp = server_list[server].chan_list; tmp; tmp = next)
		{
			next = tmp->next;
			for (Nick = tmp->nicks; Nick; Nick = Nick->next)
				add_to_whowas_buffer(Nick, tmp->channel, Nick->server, get_server_name(server));
			tmp->nicks = NULL;
			clear_bans(tmp);
			add_to_whowas_chan_buffer(tmp);
		}
		server_list[server].chan_list = NULL;
	}
	update_all_windows();
}

/*
 * remove_from_channel: removes the given nickname from the given channel. If
 * the nickname is not on the channel or the channel doesn't exist, nothing
 * happens. 
 */
void remove_from_channel(char *channel, char *nick, int server, int netsplit, char *reason)
{
	ChannelList *chan;
	NickList *tmp;
	char buf[BIG_BUFFER_SIZE+1];
	char *server1 = NULL, *server2 = NULL;

	if (netsplit && reason)
	{
		char *p = NULL;
		strncpy(buf, reason, sizeof(buf)-1);
		if ((p = strchr(buf, ' ')))
		{
			*p++ = '\0';
			server2 = buf;
			server1 = p;
		}		
	}
	if (channel)
	{
		if ((chan = lookup_channel(channel, server, CHAN_NOUNLINK)))
			if ((tmp = (NickList *) list_lookup((List **) &(chan->nicks), nick, !USE_WILDCARDS, REMOVE_FROM_LIST)))
				add_to_whowas_buffer(tmp, channel, server1 ? server1 : get_server_name(server), server2);
	}
	else
	{
		for (chan = server_list[server].chan_list; chan; chan = chan->next)
			if ((tmp = (NickList *) list_lookup((List **) &(chan->nicks), nick, !USE_WILDCARDS, REMOVE_FROM_LIST)))
				add_to_whowas_buffer(tmp, chan->channel, server1 ? server1 : get_server_name(server), server2);
	}
}

/*
 * rename_nick: in response to a changed nickname, this looks up the given
 * nickname on all you channels and changes it the new_nick 
 */
void
rename_nick(old_nick, new_nick, server)
	char	*old_nick,
		*new_nick;
	int	server;
{
	ChannelList *chan;
	NickList *tmp;

	for (chan = server_list[server].chan_list; chan; chan = chan->next)
	{
		if ((chan->server == server) != 0)
		{
			if ((tmp = (NickList *) list_lookup((List **) &chan->nicks, old_nick, !USE_WILDCARDS, !REMOVE_FROM_LIST)))
			{
				tmp->stat_nicks++;
				if ((is_other_flood(chan, tmp, NICK_FLOOD) && you_is_opped(chan->channel)) ||
					(lame_list && find_in_list((List **)&lame_list, new_nick, 0)))
					send_to_server("KICK %s %s :%s", chan->channel, new_nick, "\002Niq flood\002");	
				new_free(&tmp->nick);
				malloc_strcpy(&tmp->nick, new_nick);
			}
		}
	}
}

/*
 * is_on_channel: returns true if the given nickname is in the given channel,
 * false otherwise.  Also returns false if the given channel is not on the
 * channel list. 
 */
int
is_on_channel(channel, server, nick)
	char	*channel;
	int	server;
	char	*nick;
{
	ChannelList *chan;

	chan = lookup_channel(channel, server, CHAN_NOUNLINK);
	if (chan && chan->connected && list_lookup((List **) &(chan->nicks), nick, !USE_WILDCARDS, !REMOVE_FROM_LIST))
		return 1;
	return 0;
}

int
is_chanop(channel, nick)
	char	*channel;
	char	*nick;
{
	ChannelList *chan;
	NickList *Nick;

	if ((chan = lookup_channel(channel, from_server, CHAN_NOUNLINK)) && chan->connected)
		if ((Nick = (NickList *) list_lookup((List **)&(chan->nicks),
			nick, !USE_WILDCARDS, !REMOVE_FROM_LIST)))
				if (Nick->chanop)
					return 1;
	return 0;
}

static	void
show_channel(chan)
	ChannelList *chan;
{
	NickList *tmp;
	int	buffer_len,
		len;
	char	*nicks = NULL;
	char	*s;
	char	buffer[BIG_BUFFER_SIZE + 1];
	
	*buffer = (char) 0;
	buffer_len = 0;
	s = recreate_mode(chan);
	for (tmp = chan->nicks; tmp; tmp = tmp->next)
	{
		len = strlen(tmp->nick);
		if (buffer_len + len >= (BIG_BUFFER_SIZE / 2))
		{
			malloc_strcpy(&nicks, buffer);
			say("\t%s %c%s (%s): %s", chan->channel, s ? '+': ' ', s ? s : "<none>", get_server_name(chan->server), nicks);
			*buffer = (char) 0;
			buffer_len = 0;
		}
		strmcat(buffer, tmp->nick, BIG_BUFFER_SIZE);
		strmcat(buffer, " ", BIG_BUFFER_SIZE);
		buffer_len += len + 1;
	}
	malloc_strcpy(&nicks, buffer);
	say("\t%s %c%s (%s): %s", chan->channel, s ? '+':' ', s ? s : "<none>", get_server_name(chan->server), nicks);
	new_free(&nicks);
}

/* list_channels: displays your current channel and your channel list */
void
list_channels()
{
	ChannelList *tmp;
	int	server,
		no = 1;

	if (server_list[from_server].chan_list)
	{
		if (get_channel_by_refnum(0))
			say("Current channel %s", get_channel_by_refnum(0));
		else
			say("No current channel for this window");
		if ((tmp = server_list[get_window_server(0)].chan_list))
		{
			for (; tmp; tmp = tmp->next)
				show_channel(tmp);
			no = 0;
		}
		if (connected_to_server != 1)
		{
			for (server = 0; server < number_of_servers; server++)
			{
				if (server == get_window_server(0))
					continue;
				say("Other servers:");
				for (tmp = server_list[server].chan_list; tmp; tmp = tmp->next)
					show_channel(tmp);
				no = 0;
			}
		}
	}
	else
		say("You are not on any channels");
}

void
#ifdef __STDC__
switch_channels(char key, char *ptr)
#else
switch_channels(key, ptr)
	char	key;
	char *	ptr;
#endif
{
	ChannelList *	tmp;
	char *	s;

	if (server_list[from_server].chan_list)
	{
		if (get_channel_by_refnum(0))
		{
			if ((tmp = lookup_channel(get_channel_by_refnum(0), from_server, CHAN_NOUNLINK)))
			{
				for (tmp = tmp->next; tmp; tmp = tmp->next)
				{
					s = tmp->channel;
					if (!is_current_channel(s, from_server, 0) && !(is_bound(s, from_server) && curr_scr_win != tmp->window))
					{
						set_channel_by_refnum(0, s);
						update_all_windows();
						return;
					}
				}
			}
		}
		for (tmp = server_list[from_server].chan_list; tmp; tmp = tmp->next)
		{
			s = tmp->channel;
			if (!is_current_channel(s, from_server, 0) && !(is_bound(s, from_server) && curr_scr_win != tmp->window))
			{
				set_channel_by_refnum(0, s);
				update_all_windows();
				return;
			}
		}
	}
}

/* real_channel: returns your "real" channel (your non-multiple channel) */
char	*
real_channel()
{
	ChannelList *tmp;

	if (server_list[from_server].chan_list)
		for (tmp = server_list[from_server].chan_list; tmp; tmp = tmp->next)
			if (tmp->server == from_server && *(tmp->channel) != '#')
				return (tmp->channel);
	return (NULL);
}

void
change_server_channels(old, new)
	int	old,
		new;
{
	ChannelList *tmp;

	if (new == old)
		return;
	if (new > -1 && server_list[new].chan_list)
		server_list[new].chan_list->server = new;
	if (new > -1 && old > -1)
	{
		tmp = server_list[old].chan_list;
		server_list[new].chan_list = tmp; 
		server_list[old].chan_list = NULL;
	}
}

void
clear_channel_list(server)
	int	server;
{
	ChannelList *tmp,
		*next;
	Window		*ptr;
	int		flag = 1;

	while ((ptr = traverse_all_windows(&flag)))
		if (ptr->server == server && ptr->current_channel)
			new_free(&ptr->current_channel);
	
	for (tmp = server_list[server].chan_list; tmp; tmp = next)
	{
		NickList *tmp1, *next1;
		for (tmp1 = tmp->nicks; tmp1; tmp1 = next1)
		{
			next1 = tmp1->next;
			add_to_whowas_buffer(tmp1, tmp->channel, tmp1->server, get_server_name(server));
			
		}
		tmp->nicks = NULL;
		next = tmp->next;
		add_to_whowas_chan_buffer(tmp);
	}

	server_list[server].chan_list = NULL;
	
	return;
}

/*
 * reconnect_all_channels: used after you get disconnected from a server, 
 * clear each channel nickname list and re-JOINs each channel in the 
 * channel_list ..  
 */
void
reconnect_all_channels(server)
	int	server;
{
	ChannelList *tmp;
	char	*mode, *chan;

	for (tmp = server_list[from_server].chan_list; tmp; tmp = tmp->next)
	{
		
		mode = recreate_mode(tmp);
/* CDE possible problems here */
		if ((chan = tmp->window->current_channel) && *chan && !is_bound(chan, tmp->window->server))
			malloc_strcpy(&tmp->window->waiting_channel, chan);

#ifdef COMPAT_27
		if (get_server_version(server) >= Server2_8)
		{
#endif
			send_to_server("JOIN %s%s%s", tmp->channel, tmp->key ? " " : "", tmp->key ? tmp->key : "");
			if (NULL != mode)
				add_to_mode_list(tmp->channel, server, mode);
#ifdef COMPAT_27
		}
		else
			send_to_server("JOIN %s %s", tmp->channel, mode ? mode : "");
#endif
		clear_channel(tmp);
	}
	clear_channel_list(from_server);
	window_check_servers();
	message_from(NULL, LOG_CRAP);
}

char 	* 
channel_key(char *channel) {
        ChannelList *tmp = NULL;
        for (tmp = server_list[from_server].chan_list; tmp && strcmp(tmp->channel, channel); tmp = tmp->next)
        	;
        return tmp ? ((tmp->key)? tmp->key : empty_string) : empty_string;
}

char	*
what_channel(nick, server)
	char	*nick;
	int	server;
{
	ChannelList *tmp;

	if (curr_scr_win->current_channel && is_on_channel(curr_scr_win->current_channel, curr_scr_win->server, nick))
		return curr_scr_win->current_channel;

	for (tmp = server_list[from_server].chan_list; tmp; tmp = tmp->next)
		if (list_lookup((List **) &(tmp->nicks), nick, !USE_WILDCARDS, !REMOVE_FROM_LIST))
			return tmp->channel;

	return NULL;
}

char	*
walk_channels(nick, init, server)
	int	init;
	char	*nick;
	int	server;
{
	static	ChannelList *tmp = NULL;

	if (init)
		tmp = server_list[server].chan_list;
	else if (tmp)
		tmp = tmp->next;
	for (;tmp ; tmp = tmp->next)
		if ((tmp->server == from_server) && (list_lookup((List **) &(tmp->nicks), nick, !USE_WILDCARDS, !REMOVE_FROM_LIST)))
			return (tmp->channel);
	return NULL;
}

int
get_channel_oper(channel, server)
	char	*channel;
	int	server;
{
	ChannelList *chan;
	NickList *Nick;
	if ((chan = lookup_channel(channel, server, CHAN_NOUNLINK)))
	{
		if ((Nick = (NickList *)list_lookup((List **) &(chan->nicks), get_server_nickname(server), !USE_WILDCARDS, !REMOVE_FROM_LIST)))
			return Nick->chanop;
		else
			return 1;
	}
	else
		return 1;
}

extern	void
set_channel_window(window, channel, server)
	Window	*window;
	char	*channel;
	int	server;
{
	ChannelList	*tmp;

	if (!channel)
		return;
	for (tmp = server_list[server].chan_list; tmp; tmp = tmp->next)
		if (!my_stricmp(channel, tmp->channel))
		{
			tmp->window = window;
			return;
		}
}

extern	char	*
create_channel_list(window)
	Window	*window;
{
	ChannelList	*tmp;
	char	*value = NULL;
	char	buffer[BIG_BUFFER_SIZE + 1];
	
	*buffer = '\0';
	for (tmp = server_list[window->server].chan_list; tmp; tmp = tmp->next)
	{
		strcat(buffer, tmp->channel);
		strcat(buffer, " ");
	}
	malloc_strcpy(&value, buffer);

	return value;
}

extern void
channel_server_delete(server)
	int     server;
{
	ChannelList     *tmp;
	int	i;

	for (i = server + 1; i < number_of_servers; i++)
		for (tmp = server_list[i].chan_list ; tmp; tmp = tmp->next)
			if (tmp->server >= server)
				tmp->server--;
}

static	void
add_to_mode_list(channel, server, mode)
	char	*channel;
	int	server;
	char	*mode;
{
	struct modelist	*mptr;

	if (!channel || !*channel || !mode || !*mode)
		return;
	mptr = (struct modelist *) new_malloc(sizeof(struct modelist));
	mptr->chan = NULL;
	malloc_strcpy(&mptr->chan, channel);
	mptr->server = server;
	mptr->mode = NULL;
	malloc_strcpy(&mptr->mode, mode);
	mptr->next = mode_list;
	mode_list = mptr;
}

static	void
check_mode_list_join(channel, server)
	char	*channel;
	int	server;
{
	struct modelist *mptr = mode_list;

	for (;NULL != mptr; mptr = mptr->next)
	{
		if (0 == my_stricmp(mptr->chan, channel)
			&& mptr->server == server)
		{
			int	old_server = from_server;

			from_server = server;
			send_to_server("MODE %s %s", mptr->chan, mptr->mode);
			from_server = old_server;
			return;
		}
	}
	remove_from_mode_list(channel, server);
}

extern	void
remove_from_mode_list(channel, server)
	char	*channel;
	int	server;
{
	struct modelist	*curr, *next, 	*prev = NULL;

	for (next = mode_list; next; )
	{
		curr = next;
		next = curr->next;
		if (!my_stricmp(curr->chan, channel) && curr->server == server)
		{
			if (curr == mode_list)
				mode_list = curr->next;
			else
				prev->next = curr->next;
			prev = curr;
			new_free(&curr->chan);
			new_free(&curr->mode);
			new_free((char **)&curr);
		}
		else
			prev = curr;
	}
}

extern	int
chan_is_connected(channel, server)
	char *	channel;
	int	server;
{
	ChannelList *	cp = lookup_channel(channel, server, CHAN_NOUNLINK);

	if (!cp)
		return 0;

	return (cp->connected);
}
