/*
 * numbers.c:handles all those strange numeric response dished out by that
 * wacky, nutty program we call ircd 
 *
 *
 * written by michael sandrof
 *
 * copyright(c) 1990 
 *
 * see the copyright file, or do a help ircii copyright 
 */


#include "irc.h"

#include "input.h"
#include "edit.h"
#include "ircaux.h"
#include "vars.h"
#include "lastlog.h"
#include "list.h"
#include "hook.h"
#include "server.h"
#include "whois.h"
#include "numbers.h"
#include "window.h"
#include "screen.h"
#include "output.h"
#include "names.h"
#include "whois.h"
#include "funny.h"
#include "parse.h"
#include "misc.h"
#include "pluto.h"

static	void	reset_nickname		_((void));
static	void	nickname_in_use		_((char *, char **));
static	void	password_sendline	_((char *, char *));
static	void	get_password		_((void));
static	void	nickname_sendline	_((char *, char *));
static	void	channel_topic		_((char *, char **));
static	void	not_valid_channel	_((char *, char **));
static	void	cannot_join_channel	_((char *, char **));
static	void	version1		_((char *, char **));
static	void	invite			_((char *, char **));
static 	void	fudge_nickname		_((void));

static	int	already_doing_reset_nickname = 0;
static	int	number_of_bans = 0;

extern void	parse_364		_((char *, char *, char *));
extern void	parse_365		_((char *, char *, char *));
extern int	delay_check_auto	_((char *));
extern void	load_scripts		_((void));
extern int	stats_k_grep		_((char **));

extern int	link_look;
extern int	no_hook_notify;
extern int	in_e_nick;
extern int	who_on_join;

#define NICKNAM_LEN 9
/*
 * numeric_banner: This returns in a static string of either "xxx" where
 * xxx is the current numeric, or "***" if SHOW_NUMBERS is OFF 
 */
char	*
numeric_banner()
{
	static	char	thing[4];

	if (get_int_var(SHOW_NUMERICS_VAR))
		sprintf(thing, "%3.3u", -current_numeric);
	else
		strcpy(thing, "***");
	return (thing);
}

#if 0
int set_channel_sync(char *channel, int what)
{
ChannelList *tmp;
	if (!(tmp = lookup_channel(channel, from_server, CHAN_NOUNLINK)))
		return -1; /* channel lookup problem */
	switch(what)
	{
		case WAIT_WHO:
			tmp->flags.got_who = 1;
			break;
	}
	return 1;
}
#endif

int check_sync(int comm, char *channel, char *nick)
{
ChannelList *tmp;
BanList *new;
	if (!(tmp = lookup_channel(channel, from_server, CHAN_NOUNLINK)))
		return -1; /* channel lookup problem */
	if (!tmp->flags.got_modes && !tmp->flags.got_who && !tmp->flags.got_bans)
		return -1;	/* already synced this channel */
	
	switch (comm)
	{
		case 315: /* End of who */
		{			
			if (tmp->flags.got_who)
				tmp->flags.got_who = 0;
			break;
		}
		case 368: /* End of Bans */
		{
			if (tmp->flags.got_bans)
				tmp->flags.got_bans = 0;
			break;
		}
		case 324: /* End of channel mode */
		{
			if (tmp->flags.got_modes)
				tmp->flags.got_modes = 0;
			break;
		}
		case 367: /* bans on channel */
		{
			new = (BanList *)new_malloc(sizeof(BanList));
			malloc_strcpy(&new->ban, nick);
			new->time = time(NULL);
			add_to_list((List **)&tmp->bans,(List *)new); 
		}
		default:
			break;
	}
	
	
	if (!tmp->flags.got_modes && !tmp->flags.got_who && !tmp->flags.got_bans)
	{
		struct timeval tv;
		get_time(&tv);
		if (do_hook(CHANNEL_SYNCH_LIST, "%s %1.3f", channel, time_diff(tmp->join_time,tv)))
			bitchsay("Join to %s was synced in %1.3f secs!!", channel, time_diff(tmp->join_time,tv));
		delay_check_auto(channel);
		return 1; /*  just finished all syncs */
	}
	else
		return 0; /* not done just yet */
}

static int check_server_sync(char *from, char **ArgList)
{
static char *desync = NULL;
	if (!desync || (desync && !match(desync, from)))
	{
		if (!match(from, get_server_name(from_server)))
		{
			malloc_strcpy(&desync, from);
			if (do_hook(DESYNC_MESSAGE_LIST, "%s %s", from, ArgList[0]))
				bitchsay("%s from [%s] is Desynced from irc network", ArgList[0], from);
			return 1;
		}	 
	} 
	return 0;
}

/*
 * display_msg: handles the displaying of messages from the variety of
 * possible formats that the irc server spits out.  you'd think someone would
 * simplify this 
 */
void
display_msg(from, ArgList)
	char	*from,
		**ArgList;
{
	char	*ptr;
	char	*rest;

	rest = PasteArgs(ArgList, 0);
	if (from && (my_strnicmp(get_server_itsname(from_server), from,
			strlen(get_server_itsname(from_server))) == 0))
		from = NULL;
	if ((ptr = (char *) index(rest, ':')) != NULL)
	{
		*(ptr++) = (char) 0;
		if (strlen(rest))
		{
			if (from)
				put_it("%s %s: %s (from %s)", numeric_banner(),
					rest, ptr, from);
			else
				put_it("%s %s: %s", numeric_banner(), rest,
					ptr);
		}
		else
		{
			if (from)
				put_it("%s %s (from %s)", numeric_banner(),
					ptr, from);
			else
				put_it("%s %s", numeric_banner(), ptr);
		}
	}
	else
	{
		if (from)
			put_it("%s %s (from %s)", numeric_banner(), rest, from);
		else
			put_it("%s %s", numeric_banner(), rest);
	}
}

/*
 * password_sendline: called by send_line() in get_password() to handle
 * hitting of the return key, etc 
 */
static	void
password_sendline(data, line)
	char	*data;
	char	*line;
{
	int	new_server;

	new_server = atoi(line);
	set_server_password(new_server, line);
	connect_to_server(get_server_name(new_server),
		get_server_port(new_server), -1);
}

/*
 * get_password: when a host responds that the user needs to supply a
 * password, it gets handled here!  the user is prompted for a password and
 * then reconnection is attempted with that password.  but, the reality of
 * the situation is that no one really uses user passwords.  ah well 
 */
static	void
get_password()
{
	char	server_num[8];

	say("password required for connection to server %s",
		get_server_name(from_server));
	close_server(from_server, empty_string);
        if (!dumb)
	{
		strcpy(server_num, ltoa(from_server));
		add_wait_prompt("Server Password:", password_sendline,
			server_num, WAIT_PROMPT_LINE);
	}
}

/*ARGSUSED*/
static	void
nickname_sendline(data, nick)
	char	*data;
	char	*nick;
{
	int	new_server, server;

	new_server = atoi(data);
	if ((nick = check_nickname(nick)) != NULL)
	{
		server = from_server;
		from_server = new_server;
		send_to_server("NICK %s", nick);
		if (new_server == primary_server)
			strmcpy(nickname, nick, NICKNAME_LEN);
		set_server_nickname(new_server, nick);
		from_server = server;
		already_doing_reset_nickname = 0;
		update_all_status();
	}
	else
	{
		say("illegal nickname, try again");
		if (!dumb)
			add_wait_prompt("Nickname: ", nickname_sendline, data,
					WAIT_PROMPT_LINE);
	}
}

/*
 * reset_nickname: when the server reports that the selected nickname is not
 * a good one, it gets reset here. 
 */
static	void
reset_nickname()
{
	char	server_num[10];

	if (already_doing_reset_nickname)
		return;
	say("You have specified an illegal nickname");
	if (!dumb)
	{
		already_doing_reset_nickname = 1;
		say("Please enter your nickname");
		strcpy(server_num, ltoa(from_server));
		add_wait_prompt("Nickname: ", nickname_sendline, server_num,
			WAIT_PROMPT_LINE);
	}
	update_all_status();
}

/*ARGSUSED*/
static	void
channel_topic(from, ArgList)
	char	*from,
		**ArgList;
{
	char	*topic, *channel;

	if (ArgList[1] && is_channel(ArgList[0]))
	{
		channel = ArgList[0];
		topic = ArgList[1];
		message_from(channel, LOG_CRAP);
		put_it("%s Topic for %s: %s", numeric_banner(), channel,
			topic);
	}
	else
	{
		PasteArgs(ArgList, 0);
		message_from(NULL, LOG_CURRENT);
		put_it("%s Topic: %s", numeric_banner(), ArgList[0]);
	}
}

static	void
nickname_in_use(from, ArgList)
	char	*from,
		**ArgList;
{
	PasteArgs(ArgList, 0);
	if (is_server_connected(from_server))
		if (do_hook(current_numeric, "%s", *ArgList))
			display_msg(from, ArgList);
	else if (never_connected || from_server != primary_server ||
	    !attempting_to_connect)
	{
		if (!in_e_nick)
		{
			fudge_nickname();
			if (do_hook(current_numeric, "%s", *ArgList))
				display_msg(from, ArgList);
		}
	}
	else
	{
		send_to_server("USER %s %s %s :%s", username,
			(send_umode && *send_umode) ? send_umode : hostname,
			server_list[from_server].name, realname);
		send_to_server("NICK %s", get_server_nickname(from_server));
		if (send_umode && *send_umode == '+')
			send_to_server("MODE %s %s", get_server_nickname(from_server), send_umode);
	}
}

static	void
not_valid_channel(from, ArgList)
	char	*from,
		**ArgList;
{
	char	*channel;
	char	*s;

	if (!(channel = ArgList[0]) || !ArgList[1])
		return;
	PasteArgs(ArgList, 1);
	s = get_server_name(from_server);
	if (0 == my_strnicmp(s, from, strlen(s)))
	{
		remove_channel(channel, from_server);
		put_it("%s %s %s", numeric_banner(), channel, ArgList[1]);
	}
}

/* from ircd .../include/numeric.h */
/*
#define ERR_CHANNELISFULL    471
#define ERR_INVITEONLYCHAN   473
#define ERR_BANNEDFROMCHAN   474
#define ERR_BADCHANNELKEY    475
#define ERR_BADCHANMASK      476
*/
static	void
cannot_join_channel(from, ArgList)
	char	*from,	
		**ArgList;
{
	char	buffer[BIG_BUFFER_SIZE + 1];
	
	if (ArgList[0])
		remove_channel(ArgList[0], from_server);
	PasteArgs(ArgList, 0);
	strcpy(buffer, ArgList[0]);
	switch(-current_numeric)
	{
	case 471:
		strcat(buffer, " (Channel is full)");
		break;
	case 473:
		strcat(buffer, " (You must be invited)");
		break;
	case 474:
		strcat(buffer, " (You are banned)");
		break;
	case 475:
		strcat(buffer, " (Bad channel key)");
		break;
	case 476:
		strcat(buffer, " (Bad channel mask)");
		break;
	}
	put_it("%s %s", numeric_banner(), buffer);
}


/*ARGSUSED*/
static	void
version1(from, ArgList)
	char	*from,
		**ArgList;
{
	if (ArgList[2])
	{
		PasteArgs(ArgList, 2);
		put_it("%s Server %s: %s %s", numeric_banner(), ArgList[1],
			ArgList[0], ArgList[2]);
	}
	else
	{
		PasteArgs(ArgList, 1);
		put_it("%s Server %s: %s", numeric_banner(), ArgList[1],
			ArgList[0]);
	}
}


/*ARGSUSED*/
static	void
invite(from, ArgList)
	char	*from,
		**ArgList;
{
	char	*who,
		*channel;

	if ((who = ArgList[0]) && (channel = ArgList[1]))
	{
		message_from(channel, LOG_CRAP);
		if (do_hook(current_numeric, "%s %s %s", from, who, channel))
			put_it("%s Inviting %s to channel %s",
					numeric_banner(), who, channel);
	}
}


/*
 * numbered_command: does (hopefully) the right thing with the numbered
 * responses from the server.  I wasn't real careful to be sure I got them
 * all, but the default case should handle any I missed (sorry) 
 */
void
numbered_command(from, comm, ArgList)
	char	*from;
	int	comm;
	char	**ArgList;
{
	char	*user;
	char	none_of_these = 0;
	int	flag,
		lastlog_level;

	if (!ArgList[1] || !from || !*from)
		return;
	
	user = (*ArgList[0]) ? ArgList[0] : NULL;

	lastlog_level = set_lastlog_msg_level(LOG_CRAP);
	message_from(NULL, LOG_CRAP);
	ArgList++;
	current_numeric = -comm;	/* must be negative of numeric! */
	switch (comm)
	{
	case 001:	/* #define RPL_WELCOME          001 */
		
		set_server2_8(from_server, 1);
		load_scripts();
		PasteArgs(ArgList, 0);
		if (do_hook(current_numeric, "%s %s", from, *ArgList)) 
			display_msg(from, ArgList);
		break;
/* should do something with this some day, 2.8 had channel/user mode switches */
	case 004:	/* #define RPL_MYINFO           004 */
		got_initial_version_28(ArgList);
		load_scripts();
		PasteArgs(ArgList, 0);
		if (do_hook(current_numeric, "%s %s", from, *ArgList))
			display_msg(from, ArgList);
		break;

	case 251:
	case 252:
	case 253:
	case 254:
	case 255:
		if (do_hook(current_numeric, "%s %s", from, *ArgList))
			display_msg(from, ArgList);
		break;

	case 301:		/* #define RPL_AWAY             301 */
		user_is_away(from, ArgList);
		break;

	case 302:		/* #define RPL_USERHOST         302 */
		userhost_returned(from, ArgList);
		break;

	case 303:		/* #define RPL_ISON             303 */
		ison_returned(from, ArgList);
		break;

	case 311:		/* #define RPL_WHOISUSER        311 */
		whois_name(from, ArgList);
		break;

	case 312:		/* #define RPL_WHOISSERVER      312 */
		whois_server(from, ArgList);
		break;

	case 313:		/* #define RPL_WHOISOPERATOR    313 */
		whois_oper(from, ArgList);
		break;

	case 314:		/* #define RPL_WHOWASUSER       314 */
		whowas_name(from, ArgList);
		break;

	case 316:		/* #define RPL_WHOISCHANOP      316 */
/*		whois_chop(from, ArgList);*/
		break;

	case 317:		/* #define RPL_WHOISIDLE        317 */
		whois_lastcom(from, ArgList);
		break;

	case 318:		/* #define RPL_ENDOFWHOIS       318 */
		end_of_whois(from, ArgList);
		break;

	case 319:		/* #define RPL_WHOISCHANNELS    319 */
		whois_channels(from, ArgList);
		break;

	case 321:		/* #define RPL_LISTSTART        321 */
		ArgList[0] = "Channel\0Users\0Topic";
		ArgList[1] = ArgList[0] + 8;
		ArgList[2] = ArgList[1] + 6;
		ArgList[3] = NULL;
		funny_list(from, ArgList);
		break;

	case 322:		/* #define RPL_LIST             322 */
		funny_list(from, ArgList);
		break;

	case 324:		/* #define RPL_CHANNELMODEIS    324 */
		funny_mode(from, ArgList);
		break;

	case 341:		/* #define RPL_INVITING         341 */
		invite(from, ArgList);
		break;

	case 352:		/* #define RPL_WHOREPLY         352 */
		whoreply(NULL, ArgList);
		break;

	case 353:		/* #define RPL_NAMREPLY         353 */
		funny_namreply(from, ArgList);
		break;

	case 366:		/* #define RPL_ENDOFNAMES       366 */
	{
		int	flag = 1;
		PasteArgs(ArgList, 0);
		if (get_int_var(SHOW_END_OF_MSGS_VAR))
			flag = do_hook(current_numeric, "%s %s", from, ArgList[0]);
		
		if (!funny_is_ignore_channel())
		{
			if (get_int_var(SHOW_END_OF_MSGS_VAR) && flag)
				display_msg(from, ArgList);
		} else
			funny_set_ignore_mode();
	}
	break;

	case 381: 		/* #define RPL_YOUREOPER        381 */
		PasteArgs(ArgList, 0);
		if (!is_server_connected(from_server))
		{
			say("Odd Server stuff from %s: %s",from, ArgList[0]);
			return; 
		}
		if (do_hook(current_numeric, "%s %s", from, *ArgList))
			display_msg(from, ArgList);
		break;

	case 401:		/* #define ERR_NOSUCHNICK       401 */
		no_such_nickname(from, ArgList);
		break;

	case 421:		/* #define ERR_UNKNOWNCOMMAND   421 */
		if (check_screen_redirect(ArgList[0]))
			break;
		if (check_wait_command(ArgList[0]))
			break;
		if (!strcmp(ArgList[0], "WAITFORNOTIFY"))
		{
			no_hook_notify = 0;
			break;
		}
		PasteArgs(ArgList, 0);
		
		flag = do_hook(current_numeric, "%s %s", from, *ArgList);
		if (flag)
			display_msg(from, ArgList);
		break;

	case 432:		/* #define ERR_ERRONEUSNICKNAME 432 */
		if (do_hook(current_numeric, "%s %s", from, *ArgList))
			display_msg(from, ArgList);
		fudge_nickname();
		break;
	case 437:
	case 433:		/* #define ERR_NICKNAMEINUSE    433 */ 
		nickname_in_use(from, ArgList);
		fudge_nickname();
		break;

	case 464:		/* #define ERR_PASSWDMISMATCH   464 */
		PasteArgs(ArgList, 0);
		if (!is_server_connected(from_server))
		{
			say("Odd Server stuff from %s: %s",from, ArgList[0]);
			return; 
		}
		flag = do_hook(current_numeric, "%s %s", from, ArgList[0]);
		if (oper_command)
		{
			if (flag)
				display_msg(from, ArgList);
		}
		else
			get_password();
		break;

	case 465:		/* #define ERR_YOUREBANNEDCREEP 465 */
	{

		PasteArgs(ArgList, 0);
		if (!is_server_connected(from_server))
			return;
		if (do_hook(current_numeric, "%s %s", from, ArgList[0]))
			display_msg(from, ArgList);
				
		close_server(from_server, empty_string);
		window_check_servers();
		if (from_server == primary_server)
			get_connected(from_server + 1);
		break;
	}

	case 367:
		if (check_sync(comm, ArgList[0], ArgList[1]) == 0)
			break;
		if (ArgList[2])
		{
			time_t tme = (time_t) strtoul(ArgList[3], NULL, 10);
			if (do_hook(current_numeric, "%s %s %s %s %s", 
				from, ArgList[0], ArgList[1], ArgList[2], ArgList[3]))
			put_it("%s %s %-25s set by %-10s %lu sec ago", 
				numeric_banner(), ArgList[0],
				ArgList[1], ArgList[2], 
				(unsigned long)(time(NULL) - tme));
		}
		else
			if (do_hook(current_numeric, "%s %s %s", from, ArgList[0], ArgList[1]))
				put_it("%s %s %s", numeric_banner(), ArgList[0], ArgList[1]);
		break;
	case 368:		/* #define RPL_ENDOFBANLIST     368 */
	{
		if (check_sync(comm, ArgList[0], NULL) == 0)
			break;
		if (get_int_var(SHOW_END_OF_MSGS_VAR))
		{
			if (do_hook(current_numeric, "%s %d %s", from, number_of_bans, *ArgList))
			{
				put_it("%s Total Number of Bans on %s - %d",
					numeric_banner(), ArgList[0],
					number_of_bans);
			}
		}
		number_of_bans = 0;
		break;
	}
	case 364:
		if (comm == 364 && get_int_var(LLOOK_VAR) && link_look) 
		{
			parse_364(ArgList[0], ArgList[1], ArgList[2] ? ArgList[2] : "0");
			break;
		}
	case 365:		/* #define RPL_ENDOFLINKS       365 */
		if (comm == 365 && get_int_var(LLOOK_VAR) && link_look) {
			parse_365(ArgList[0], ArgList[1],NULL);
			link_look--;
			break;
		}
	case 216: /* STATS k */
		if (comm == 216 && stats_k_grep(ArgList))
			break;
	case 404:
	case 442:
	case 482:
		if (comm == 404 || comm == 482 || comm == 442)
		{
			if (check_server_sync(from, ArgList))
			{
				send_to_server("WHO -server %s %s", from, ArgList[0]);
				break;
			}
		}
	/*
	 * The following accumulates the remaining arguments
	 * in ArgSpace for hook detection. We can't use
	 * PasteArgs here because we still need the arguments
	 * separated for use elsewhere.
	 */
	default:
		{
			char	*ArgSpace = NULL;
			int	i,
				len,
				do_message_from = 0;

			for (i = len = 0; ArgList[i]; len += strlen(ArgList[i++]))
				;
			len += (i - 1);
			ArgSpace = new_malloc(len + 1);
			ArgSpace[0] = '\0';
			/* this is cheating */

			if (ArgList[0] && is_channel(ArgList[0]))
				do_message_from = 1;
			for (i = 0; ArgList[i]; i++)
			{
				if (i)
					strcat(ArgSpace, " ");
				strcat(ArgSpace, ArgList[i]);
			}
			if (do_message_from)
				message_from(ArgList[0], LOG_CRAP);
			if (!do_hook(current_numeric, "%s %s", from, ArgSpace))
			{
				new_free(&ArgSpace);
				if (do_message_from)
					message_from(NULL, lastlog_level);
				return;
			}
			if (do_message_from)
				message_from(NULL, lastlog_level);
			new_free(&ArgSpace);
			none_of_these = 1;
		}
	}
	/* the following do not hurt the ircII if intercepted by a hook */
	if (none_of_these)
	{
		switch (comm)
		{
		case 221: 		/* #define RPL_UMODEIS          221 */
			put_it("%s Your user mode is \"%s\"", numeric_banner(),
				ArgList[0]);
			break;

		case 242:		/* #define RPL_STATSUPTIME      242 */
			PasteArgs(ArgList, 0);
			if (from && !my_strnicmp(get_server_itsname(from_server),
			    from, strlen(get_server_itsname(from_server))))
				from = NULL;
			if (from)
				put_it("%s %s from (%s)", numeric_banner(),
					*ArgList, from);
			else
				put_it("%s %s", numeric_banner(), *ArgList);
			break;

		case 272:/* ENDOFSILENCE */
			PasteArgs(ArgList, 0);
			put_it("%s %s", numeric_banner(), ArgList[0]);
			break;
		case 332:		/* #define RPL_TOPIC            332 */
			channel_topic(from, ArgList);
			break;

		case 351:		/* #define RPL_VERSION          351 */
			version1(from, ArgList);
			break;

		case 364:		/* #define RPL_LINKS            364 */
		{

			if (ArgList[2])
			{
				PasteArgs(ArgList, 2);
				put_it("%s %-20s %-20s %s", numeric_banner(),
					ArgList[0], ArgList[1], ArgList[2]);
			}
			else
			{
				PasteArgs(ArgList, 1);
				put_it("%s %-20s %s", numeric_banner(),
					ArgList[0], ArgList[1]);
			}
		}
		break;

		case 372:		/* #define RPL_MOTD             372 */
			if (!get_int_var(SUPPRESS_SERVER_MOTD_VAR) ||
			    !get_server_motd(from_server))
			{
				PasteArgs(ArgList, 0);
				put_it("%s %s", numeric_banner(), ArgList[0]);
			}
			break;

		case 375:		/* #define RPL_MOTDSTART        375 */
			if (!get_int_var(SUPPRESS_SERVER_MOTD_VAR) ||
			    !get_server_motd(from_server))
			{
				PasteArgs(ArgList, 0);
				put_it("%s %s", numeric_banner(), ArgList[0]);
			}
			break;

		case 376:		/* #define RPL_ENDOFMOTD        376 */
			if (get_int_var(SHOW_END_OF_MSGS_VAR) &&
			    (!get_int_var(SUPPRESS_SERVER_MOTD_VAR) ||
			    !get_server_motd(from_server)))
			{
				PasteArgs(ArgList, 0);
				put_it("%s %s", numeric_banner(), ArgList[0]);
			}
			set_server_motd(from_server, 0);
			break;

		case 384:		/* #define RPL_MYPORTIS         384 */
			PasteArgs(ArgList, 0);
			put_it("%s %s %s", numeric_banner(), ArgList[0], user);
			break;

		case 385:		/* #define RPL_NOTOPERANYMORE   385 */
			set_server_operator(from_server, 0);
			display_msg(from, ArgList);
			update_all_status();
			break;

		case 403:		/* #define ERR_NOSUCHCHANNEL    403 */
			not_valid_channel(from, ArgList);
			break;

		case 432:		/* #define ERR_ERRONEUSNICKNAME 432 */
			display_msg(from, ArgList);
			reset_nickname();
			break;

		case 451:		/* #define ERR_NOTREGISTERED    451 */
	/*
	 * Sometimes the server doesn't catch the USER line, so
	 * here we send a simplified version again  -lynx 
	 */
			send_to_server("USER %s %s %s :%s", username,
				(send_umode && *send_umode) ? send_umode : hostname,
				server_list[from_server].name, realname);
			send_to_server("NICK %s",
				get_server_nickname(from_server));
			break;

		case 462:		/* #define ERR_ALREADYREGISTRED 462 */
			send_to_server("NICK %s",
				get_server_nickname(from_server));
			break;

		case 471:		/* #define ERR_CHANNELISFULL    471 */
		case 473:		/* #define ERR_INVITEONLYCHAN   473 */
		case 474:		/* #define ERR_BANNEDFROMCHAN   474 */
		case 475: 		/* #define ERR_BADCHANNELKEY    475 */
		case 476:		/* #define ERR_BADCHANMASK      476 */
			cannot_join_channel(from, ArgList);
			break;

#define RPL_CLOSEEND         363
#define RPL_SERVLISTEND      235
		case 315:		/* #define RPL_ENDOFWHO         315 */
			{
				if (check_sync(comm, ArgList[0], NULL) == 0)
					break;
				if (who_on_join)
				{
					who_on_join--;
					break;
				}
			}
		case 323:               /* #define RPL_LISTEND          323 */
			funny_print_widelist();

		case 219:		/* #define RPL_ENDOFSTATS       219 */
			if (comm == 219)
				stats_k_grep_end();
		case 232:		/* #define RPL_ENDOFSERVICES    232 */
		case 365:		/* #define RPL_ENDOFLINKS       365 */
			if (comm == 365 && get_int_var(LLOOK_VAR) && link_look) {
				parse_365(ArgList[0], ArgList[1],NULL);
				link_look--;
				break;
			}
		case 369:		/* #define RPL_ENDOFWHOWAS      369 */
		case 374:		/* #define RPL_ENDOFINFO        374 */
		case 394:		/* #define RPL_ENDOFUSERS       394 */
			if (get_int_var(SHOW_END_OF_MSGS_VAR))
			{
				PasteArgs(ArgList, 0);
				if (do_hook(current_numeric, "%s %s", from, *ArgList))
					display_msg(from, ArgList);
			}
			break;
		default:
			display_msg(from, ArgList);
		}
	}
	set_lastlog_msg_level(lastlog_level);
}

/* Dont even think of calling this function with anything that isnt
 * really a nickname -- it may cause segfaults!
 */
/* This will generate up to 18 nicknames plus the 9-length(nickname)
 * that are unique but still have some semblance of the original.
 * This is intended to allow the user to get signed back on to
 * irc after a nick collision without their having to manually
 * type a new nick every time..
 * 
 * The func will try to make an intelligent guess as to when it is
 * out of guesses, and if it ever gets to that point, it will do the
 * manually-ask-you-for-a-new-nickname thing.
 */
static void fudge_nickname _((void))
{
	static char oldnick[NICKNAME_LEN+1] = {0};
	static int fudge_index = 0;
	char blah[4];
	extern int from_server;
	char nickname[NICKNAME_LEN + 1] = {0};

	if (in_e_nick)
		return;		/* if we're changing nickname, dont bother */

	strcpy(nickname, get_server_nickname(from_server));

	/* First time through... */
	if (!*oldnick)
	{
		strncpy(oldnick, nickname, NICKNAM_LEN);
		fudge_index = strlen(nickname);
	}
	else
	{
		/* Same nick as the last one we convoluted? */
		if (!my_stricmp(oldnick, nickname))
		{
			fudge_index++;
			if (fudge_index == 17)
			{
				/* give up... */
				reset_nickname();
				fudge_index = 0;
				return;
			}
		}
		/* Different nick then last time.. start over. */
		else
		{
			strcpy(oldnick, nickname);
			fudge_index = strlen(nickname);
		}
	}

	/* 
	 * Process of fudging a nickname:
	 * If the nickname length is less then 9, add an underscore.
	 */
	if (strlen(nickname) < 9)
		strcat(nickname, "_");

	/* 
	 * The nickname is 9 characters long. roll the nickname
	 */
	else
	{
		char tmp = nickname[8];
		nickname[8] = nickname[7];
		nickname[7] = nickname[6];
		nickname[6] = nickname[5];
		nickname[5] = nickname[4];
		nickname[4] = nickname[3];
		nickname[3] = nickname[2];
		nickname[2] = nickname[1];
		nickname[1] = nickname[0];
		nickname[0] = tmp;
	}

	strcpy(blah, ltoa(from_server));
	strcpy(oldnick, nickname);
	nickname_sendline(blah, nickname);
}
